@@ -42,6 +42,7 @@ Now you can perform operations against it using this module:
 
 """
 
+import string
 import cPickle, os, socket, stat
 try:
     import fcntl
@@ -52,8 +53,14 @@ try:
 except ImportError:
     errno = None
 
+try:
+    import webbrowser
+except ImportError:
+    webbrowser = None
+
 from types import StringType
 from warnings import warn, filterwarnings
+from urllib import quote_plus
 
 # Change the default warning filtering for this module
 filterwarnings('always', module=__name__)
@@ -64,15 +71,24 @@ __all__ = ['NwsServer', 'NetWorkSpace', 'NwsValueIterator',
            'NwsUnsupportedOperationException', 'NwsOperationException',
            'NwsNoWorkSpaceException', 'NwsDeclarationFailedException',
            'NwsPicklingException', 'NwsUnpicklingException',
+           'NwsUnsupportedMethodException',
            'FIFO', 'LIFO', 'MULTI', 'SINGLE',
            'STRING', 'DICT',
+           'OPT_SSLKEY', 'OPT_SSLCERT', 'OPT_SSL', 'OPT_DEADMAN', 'OPT_METADATATOSERVER', 'OPT_METADATAFROMSERVER',
            'WS_NAME', 'WS_MINE', 'WS_OWNER', 'WS_PERSISTENT', 'WS_NUMVARS', 'WS_VARLIST',
-           'V_VARIABLE', 'V_VALUES', 'V_FETCHERS', 'V_FINDERS', 'V_MODE']
+           'V_VARIABLE', 'V_VALUES', 'V_FETCHERS', 'V_FINDERS', 'V_MODE',
+           'PROTOCOL_VERSION_OLD', 'PROTOCOL_VERSION_NEW_0']
 
 _PythonFP =     0x01000000
 _DirectString = 0x00000001
-_Handshake =    '1112'
 _OldProtocol =  '2222'
+_NegotiationHandshakeInit      = 'X000'
+_NegotiationHandshakeAdvertise = 'P000'
+_NegotiationHandshakeRequest   = 'R000'
+_NegotiationHandshakeAccept    = 'A000'
+
+PROTOCOL_VERSION_OLD   = 'old'
+PROTOCOL_VERSION_NEW_0 = 'new-0000'
 
 NWS_PORT = 8765
 
@@ -84,6 +100,13 @@ SINGLE = 'single'
 STRING = 'string'
 DICT = 'dict'
 
+OPT_SSLKEY             = "SSLKey"
+OPT_SSLCERT            = "SSLCertificate"
+OPT_SSL                = "SSL"
+OPT_DEADMAN            = "KillServerOnClose"
+OPT_METADATATOSERVER   = "MetadataToServer"
+OPT_METADATAFROMSERVER = "MetadataFromServer"
+
 WS_NAME = 0
 WS_MINE = 1
 WS_OWNER = 2
@@ -100,6 +123,20 @@ V_MODE = 4
 _modes = [FIFO, LIFO, MULTI, SINGLE]
 _formats = [STRING, DICT]
 
+def _marshaldict(d):
+    if d is None:
+        return "0000"
+    else:
+        n = len(d)
+        return ("%04d" % n) + ''.join(["%04d%s%04d%s" % (len(k), k, len(v), v) for k, v in d.items()])
+
+def _marshallist(d, extraArgs=0):
+    if d is None:
+        return "%04d" % extraArgs
+    else:
+        n = len(d) + extraArgs
+        return ("%04d" % n) + ''.join(["%020d%s" % (len(k), k) for k in d])
+
 def _getintattr(obj, attr, defval=0):
     try:
         val = getattr(obj, attr)
@@ -166,6 +203,9 @@ class NwsNoWorkSpaceException(NwsOperationException):
 class NwsDeclarationFailedException(NwsOperationException):
     """Variable declaration failed."""
 
+class NwsUnsupportedMethodException(NwsException):
+    """Unsupported method."""
+
 class NwsPicklingException(NwsOperationException):
     """Error pickling an object."""
 
@@ -197,6 +237,27 @@ class NwsUnpicklingException(NwsOperationException):
         return "caught %s while unpickling object retrieved from '%s @ %s'" % \
                 (repr(self._exname), self._wsName, self._varName)
 
+class SslSocket(object):
+    """Simple wrapper class giving an SSL object a minimal socket interface
+    allowing seamless switching between plain sockets and SSL.
+    """
+
+    def __init__(self, sock, keyfile=None, certfile=None):
+        self.sock   = sock
+        self.sslobj = socket.ssl(sock, keyfile, certfile)
+
+    def recv(self, len):
+        return self.sslobj.read(len)
+
+    def sendall(self, bytes):
+        return self.sslobj.write(bytes)
+
+    def shutdown(self, n):
+        self.sock.shutdown(n)
+
+    def close(self):
+        self.sock.close()
+
 class NwsServer(object):
 
     """Perform operations against the NWS server.
@@ -205,7 +266,7 @@ class NwsServer(object):
     by using the NetWorkSpace class.
     """
 
-    def __init__(self, host='localhost', port=NWS_PORT):
+    def __init__(self, host='localhost', port=NWS_PORT, connopts=None):
         """Create a connection to the NWS server.
 
         Arguments:
@@ -217,9 +278,45 @@ class NwsServer(object):
         """
         self.serverHost = host
         self.serverPort = port
+        if connopts is not None:
+            self.connopts = dict(connopts)
+        else:
+            self.connopts = None
+        self.sslkey  = None
+        self.sslcert = None
+        self.lastReceivedMetadata = None
+        if self.connopts is not None:
+            if self.connopts.has_key(OPT_SSLKEY):
+                self.sslkey  = self.connopts[OPT_SSLKEY]
+                del self.connopts[OPT_SSLKEY]
+            if self.connopts.has_key(OPT_SSLCERT):
+                self.sslcert = self.connopts[OPT_SSLCERT]
+                del self.connopts[OPT_SSLCERT]
+        if (self.sslkey is None) != (self.sslcert is None):
+            raise NwsConnectException("""invalid SSL configuration while connecting to the NWS server at %s:%d
+  Either both or none of SSLKey and SSLCertificate must be specified in the connection options.""" % \
+                    (self.serverHost, self.serverPort))
         self._connect()
 
+    def receive_name(self):
+        len = int(self._recvN(4))
+        return self._recvN(len)
+
+    def receive_dict(self):
+        num_opts = int(self._recvN(4))
+        advertise_opts = {}
+        for i in range(0, num_opts):
+            name = self.receive_name()
+            value = self.receive_name()
+            advertise_opts[name] = value
+        return advertise_opts
+
     def _connect(self):
+        if self.sslkey is not None:
+            if self.connopts is None:
+                self.connopts = dict()
+            self.connopts[OPT_SSL] = ""
+
         self.nwsSocket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
 
         try:
@@ -241,9 +338,28 @@ class NwsServer(object):
             fcntl.fcntl(fd, fcntl.F_SETFD, f)
 
         # tell the server that we're a new client
-        self._sendAll(_Handshake)
+        self._sendAll(_NegotiationHandshakeInit)
         self._handshake = self._recvN(4)
-        self._cookieProtocol = self._handshake != _OldProtocol
+        if self._handshake == _NegotiationHandshakeAdvertise:
+            self._protocolVersion = PROTOCOL_VERSION_NEW_0
+            self._cookieProtocol = True
+            advertise_opts = self.receive_dict()
+            if self.connopts is None and len(advertise_opts) != 0:
+                self.connopts = {}
+            for k, v in self.connopts.items():
+                if not advertise_opts.has_key(k):
+                    raise NwsConnectException('requested connection option %s="%s" but this server does not support this option' % (k, v))
+                elif advertise_opts[k] != '' and advertise_opts[k] != v:
+                    raise NwsConnectException('requested connection option %s="%s" but this server requires %s="%s"' % (k, v, k, advertise_opts[k]))
+            self._sendAll(_NegotiationHandshakeRequest + _marshaldict(self.connopts))
+            if self._recvN(4) != _NegotiationHandshakeAccept:
+                raise NwsConnectException('failed options negotiation with NWS server at %s:%d' % \
+                           (self.serverHost, self.serverPort))
+            if self.connopts is not None and self.connopts.has_key(OPT_SSL):
+                self.nwsSocket = SslSocket(self.nwsSocket, self.sslkey, self.sslcert)
+        else:
+            self._protocolVersion = PROTOCOL_VERSION_OLD
+            self._cookieProtocol = self._handshake != _OldProtocol
 
     def __getstate__(self):
         return {'serverHost': self.serverHost, 'serverPort': self.serverPort}
@@ -255,6 +371,9 @@ class NwsServer(object):
     def __str__(self):
         return "NwsServer %s:%d" % (self.serverHost, self.serverPort)
 
+    def get_protocol_version(self):
+        return self._protocolVersion
+
     def _recvN(self, n, fobj=None):
         if n > 0:
             if fobj is not None:
@@ -299,6 +418,27 @@ class NwsServer(object):
     def _sendAll(self, b):
         return self.nwsSocket.sendall(b)
 
+    def sendOp(self, *args, **metadata):
+        if self.connopts.has_key(OPT_METADATATOSERVER):
+            self._sendAll(_marshaldict(metadata) + _marshallist(args))
+        else:
+            self._sendAll(_marshallist(args))
+
+    def sendOpStreaming(self, *args, **metadata):
+        if self.connopts.has_key(OPT_METADATATOSERVER):
+            self._sendAll(_marshaldict(metadata) + _marshallist(args, 1))
+        else:
+            self._sendAll(_marshallist(args, 1))
+
+    def receiveMetadata(self):
+        if self.connopts.has_key(OPT_METADATAFROMSERVER):
+            return self.receive_dict()
+        else:
+            return None
+
+    def _receiveMetadata(self):
+        self.lastReceivedMetadata = self.receiveMetadata()
+
     def close(self):
         """Close the connection to the NWS server.
 
@@ -315,7 +455,7 @@ class NwsServer(object):
             self.nwsSocket.close()
         except: pass
 
-    def deleteWs(self, wsName):
+    def deleteWs(self, wsName, metadata={}):
         """Delete the specfied workspace on the NWS server.
 
         s.deleteWs(wsName)
@@ -326,15 +466,12 @@ class NwsServer(object):
 
         """
         _sanityStrObj(wsName, 'workspace name')
-
-        op = 'delete ws'
-
-        self._sendAll('0002%020d%s%020d%s' % (len(op), op, len(wsName), wsName))
-
+        self.sendOp('delete ws', wsName, **metadata)
+        self._receiveMetadata()
         status = int(self._recvN(4))
         if status: raise NwsOperationException('deleteWs failed')
 
-    def listWss(self, format=STRING):
+    def listWss(self, format=STRING, metadata={}):
         """Return a listing of all of the workspaces on the NWS server.
 
         s.listWss([format]) -> string or dictionary
@@ -351,8 +488,8 @@ class NwsServer(object):
         if format not in _formats:
             raise ValueError('illegal format: ' + format)
 
-        op = 'list wss'
-        self._sendAll('0001%020d%s' % (len(op), op))
+        self.sendOp('list wss', **metadata)
+        self._receiveMetadata()
         status = int(self._recvN(4))
         desc = self._recvN(20) # unused at the moment.
         if self._cookieProtocol:
@@ -378,7 +515,7 @@ class NwsServer(object):
 
         return wss
 
-    def mktempWs(self, wsName='__pyws__%d'):
+    def mktempWs(self, wsName='__pyws__%d', metadata={}):
         """Make a temporary workspace on the NWS server.
 
         s.mktempWs([wsName]) -> string
@@ -413,8 +550,8 @@ class NwsServer(object):
         """
         _sanityStrObj(wsName, 'workspace name')
 
-        op = 'mktemp ws'
-        self._sendAll('0002%020d%s%020d%s' % (len(op), op, len(wsName), wsName))
+        self.sendOp('mktemp ws', wsName, **metadata)
+        self._receiveMetadata()
         status = int(self._recvN(4))
         desc = self._recvN(20) # unused at the moment.
         if self._cookieProtocol:
@@ -424,7 +561,7 @@ class NwsServer(object):
         if status: raise NwsOperationException('mktempWs failed')
         return name
 
-    def openWs(self, wsName, space=None, **opt):
+    def openWs(self, wsName, space=None, metadata={}, **opt):
         """Open a workspace.
 
         s.openWs(wsName[, space]) -> space
@@ -495,28 +632,23 @@ class NwsServer(object):
             space = NetWorkSpace(wsName, server=self)
         elif space.currentWs() != wsName:
             raise ValueError('name of the specified workspace is incorrect')
-
-        op = 'open ws'
-        owner = '%d' % os.getpid()
-
-        p = 'no'
-        if opt.get('persistent', False): p = 'yes'
-
-        if opt.get('create', True):
-            self._sendAll('0004%020d%s%020d%s%020d%s%020d%s' % \
-                    (len(op), op, len(wsName), wsName, len(owner), owner,
-                        len(p), p))
         else:
-            create = 'no'
-            self._sendAll('0005%020d%s%020d%s%020d%s%020d%s%020d%s' % \
-                    (len(op), op, len(wsName), wsName, len(owner), owner,
-                        len(p), p, len(create), create))
+            owner = '%d' % os.getpid()
 
-        status = int(self._recvN(4))
-        if status: raise NwsNoWorkSpaceException("workspace %s doesn't exist" % repr(wsName))
+            p = 'no'
+            if opt.get('persistent', False): p = 'yes'
+
+            if opt.get('create', True):
+                self.sendOp('open ws', wsName, owner, p, **metadata)
+            else:
+                self.sendOp('open ws', wsName, owner, p, 'no', **metadata)
+
+            self._receiveMetadata()
+            status = int(self._recvN(4))
+            if status: raise NwsNoWorkSpaceException("workspace %s doesn't exist" % repr(wsName))
         return space
 
-    def useWs(self, wsName, space=None, **opt):
+    def useWs(self, wsName, space=None, metadata={}, **opt):
         """Use a NetWorkSpace object.
 
         s.useWs(wsName[, space]) -> space
@@ -582,26 +714,29 @@ class NwsServer(object):
             space = NetWorkSpace(wsName, server=self)
         elif space.currentWs() != wsName:
             raise ValueError('name of the specified workspace is incorrect')
-
-        op = 'use ws'
-        owner = ''
-
-        p = 'no'
-
-        if opt.get('create', True):
-            self._sendAll('0004%020d%s%020d%s%020d%s%020d%s' % \
-                    (len(op), op, len(wsName), wsName, len(owner), owner,
-                        len(p), p))
         else:
-            create = 'no'
-            self._sendAll('0005%020d%s%020d%s%020d%s%020d%s%020d%s' % \
-                    (len(op), op, len(wsName), wsName, len(owner), owner,
-                        len(p), p, len(create), create))
+            owner = ''
 
-        status = int(self._recvN(4))
-        if status: raise NwsNoWorkSpaceException("workspace %s doesn't exist" % repr(wsName))
+            p = 'no'
+
+            if opt.get('create', True):
+                self.sendOp('use ws', wsName, owner, p, **metadata)
+            else:
+                self.sendOp('use ws', wsName, owner, p, 'no', **metadata)
+
+            self._receiveMetadata()
+            status = int(self._recvN(4))
+            if status: raise NwsNoWorkSpaceException("workspace %s doesn't exist" % repr(wsName))
         return space
 
+    def view(self, port=8766):
+        if webbrowser:
+            vstring = "http://%s:%d/doit?op=listWss" % \
+                    (self.serverHost, port)
+            webbrowser.open_new(vstring)
+        else:
+            raise NwsUnsupportedMethodException('cannot import webbrowser module')
+
 class NetWorkSpace(object):
 
     """Perform operations against workspaces on NWS servers.
@@ -628,7 +763,7 @@ class NetWorkSpace(object):
     returning a default value instead.
     """
 
-    def __init__(self, wsName='__default', serverHost='localhost', serverPort=NWS_PORT, useUse=False, server=None, **opt):
+    def __init__(self, wsName='__default', serverHost='localhost', serverPort=NWS_PORT, useUse=False, server=None, connopts=None, wsmetadata={}, **opt):
         """Construct a NetWorkSpace object for the specified NwsServer.
 
         Arguments:
@@ -658,12 +793,12 @@ class NetWorkSpace(object):
                 The default value is False, meaning you are willing to
                 take ownership of this workspace.
 
-        server -- NwsServer object to associate with this object.  If
-                the value is None (the default value), then a NwsServer
-                object will be constructed, using the host and port
-                specified with the serverHost and serverPort arguments.
+        server -- This argument is only intended for internal use.
+                Specifying a value other than None will result in a
+                NetWorkSpace object that isn't properly initialized.
 
-                The default value is None.
+        connopts -- connection options for options negotiation with the
+                server.
 
         Keyword Arguments:
 
@@ -699,25 +834,32 @@ class NetWorkSpace(object):
         self._useUse = useUse
         self._opt = opt
 
+        self.lastReceivedMetadata = None
+        if len(wsmetadata) != 0:
+            if connopts is None:
+                connopts = {}
+            connopts[OPT_METADATATOSERVER] = '1'
+
         # if invoked (indirectly) via a NwsServer openWs or useWs method,
         # the server will be passed in and used. if invoked directly,
         # need to create a new NwsServer instance.
         if not server:
-            self.server = NwsServer(serverHost, serverPort)
-            # now give the server a chance to do its thing.
-            try:
-                if self._useUse:
-                    self.server.useWs(self.curWs, self, **self._opt)
-                else:
-                    self.server.openWs(self.curWs, self, **self._opt)
-            except Exception, e:
-                # close the server and re-raise the exception
-                try: self.server.close()
-                except: pass
-                raise e
+            self.server = NwsServer(serverHost, serverPort, connopts=connopts)
         else:
             self.server = server
 
+        # now give the server a chance to do its thing.
+        try:
+            if self._useUse:
+                self.server.useWs(self.curWs, self, metadata=wsmetadata, **self._opt)
+            else:
+                self.server.openWs(self.curWs, self, metadata=wsmetadata, **self._opt)
+        except Exception, e:
+            # close the server and re-raise the exception
+            try: self.server.close()
+            except: pass
+            raise e
+
         self.send = self.server._sendAll
         self.recv = self.server._recvN
         self._handshake = self.server._handshake
@@ -749,6 +891,9 @@ class NetWorkSpace(object):
     def __str__(self):
         return "NetWorkspace '%s' [%s]" % (self.curWs, str(self.server))
 
+    def receiveMetadata(self):
+        self.lastReceivedMetadata = self.server.receiveMetadata()
+
     def currentWs(self):
         """Return the name of the current workspace.
 
@@ -757,7 +902,7 @@ class NetWorkSpace(object):
         """
         return self.curWs
 
-    def declare(self, varName, mode):
+    def declare(self, varName, mode, metadata={}):
         """Declare a variable in a workspace with the specified mode.
 
         ws.declare(varName, mode)
@@ -798,13 +943,12 @@ class NetWorkSpace(object):
         if mode not in _modes and not mode.startswith('__'):
             raise ValueError('illegal mode: ' + str(mode))
 
-        op = 'declare var'
-        self.send('0004%020d%s%020d%s%020d%s%020d%s' % (len(op), op, len(self.curWs), self.curWs, len(varName), varName, len(mode), mode))
-
+        self.server.sendOp('declare var', self.curWs, varName, mode, **metadata)
+        self.receiveMetadata()
         status = int(self.recv(4))
         if status: raise NwsDeclarationFailedException('variable declaration failed')
 
-    def deleteVar(self, varName):
+    def deleteVar(self, varName, metadata={}):
         """Delete a variable from a workspace.
 
         ws.deleteVar(varName)
@@ -819,20 +963,18 @@ class NetWorkSpace(object):
         """
         _sanityStrObj(varName, 'variable name')
 
-        op = 'delete var'
-
-        self.send('0003%020d%s%020d%s%020d%s' % (len(op), op, len(self.curWs), self.curWs, len(varName), varName))
+        self.server.sendOp('delete var', self.curWs, varName, **metadata)
 
+        self.receiveMetadata()
         status = int(self.recv(4))
         if status: raise NwsOperationException('deleteVar failed')
 
-    def __retrieve(self, varName, op, missing=None, fobj=None):
+    def __retrieve(self, varName, op, missing=None, fobj=None, metadata={}):
         _sanityStrObj(varName, 'variable name')
         _sanityFileObj(fobj, 'w')
 
-        self.send('0003%020d%s%020d%s%020d%s' % \
-                (len(op), op, len(self.curWs), self.curWs,
-                 len(varName), varName))
+        self.server.sendOp(op, self.curWs, varName, **metadata)
+        self.receiveMetadata()
         status = int(self.recv(4)) # barely used at the moment.
 
         # even if failure status, read the rest of the bytes.
@@ -858,7 +1000,7 @@ class NetWorkSpace(object):
         else:
             return n
 
-    def fetch(self, varName):
+    def fetch(self, varName, metadata={}):
         """Return and remove a value of a variable from a workspace.
 
         ws.fetch(varName) -> object
@@ -878,9 +1020,9 @@ class NetWorkSpace(object):
         varName -- Name of the workspace variable to fetch.
 
         """
-        return self.__retrieve(varName, 'fetch', None)
+        return self.__retrieve(varName, 'fetch', None, metadata=metadata)
 
-    def fetchTry(self, varName, missing=None):
+    def fetchTry(self, varName, missing=None, metadata={}):
         """Try to return and remove a value of a variable from a workspace.
 
         ws.fetchTry(varName[, missing]) -> object
@@ -902,13 +1044,13 @@ class NetWorkSpace(object):
 
         """
         try:
-            return self.__retrieve(varName, 'fetchTry', missing)
+            return self.__retrieve(varName, 'fetchTry', missing, metadata=metadata)
         except NwsUnpicklingException:
             raise
         except NwsOperationException:
             return missing
 
-    def find(self, varName):
+    def find(self, varName, metadata={}):
         """Return a value of a variable from a workspace.
 
         ws.find(varName) -> object
@@ -925,9 +1067,9 @@ class NetWorkSpace(object):
         varName -- Name of the workspace variable to find.
 
         """
-        return self.__retrieve(varName, 'find', None)
+        return self.__retrieve(varName, 'find', None, metadata=metadata)
 
-    def findTry(self, varName, missing=None):
+    def findTry(self, varName, missing=None, metadata={}):
         """Try to return a value of a variable in the workspace.
 
         ws.findTry(varName[, missing]) -> object
@@ -948,13 +1090,13 @@ class NetWorkSpace(object):
 
         """
         try:
-            return self.__retrieve(varName, 'findTry', missing)
+            return self.__retrieve(varName, 'findTry', missing, metadata=metadata)
         except NwsUnpicklingException:
             raise
         except NwsOperationException:
             return missing
 
-    def listVars(self, wsName=None, format=STRING):
+    def listVars(self, wsName=None, format=STRING, metadata={}):
         """Return a listing of the variables in the workspace.
 
         ws.listVars([wsName[, format]]) -> string or dictionary
@@ -983,9 +1125,9 @@ class NetWorkSpace(object):
         if format not in _formats:
             raise ValueError('illegal format: ' + format)
 
-        op = 'list vars'
         if not wsName: wsName = self.curWs
-        self.send('0002%020d%s%020d%s' % (len(op), op, len(wsName), wsName))
+        self.server.sendOp('list vars', wsName, **metadata)
+        self.receiveMetadata()
         status = int(self.recv(4))
         desc = self.recv(20) # unused at the moment.
         if self.server._cookieProtocol:
@@ -1008,7 +1150,7 @@ class NetWorkSpace(object):
 
         return vars
 
-    def store(self, varName, val):
+    def store(self, varName, val, metadata={}):
         """Store a new value into a variable in the workspace.
 
         ws.store(varName, val)
@@ -1022,8 +1164,6 @@ class NetWorkSpace(object):
         """
         _sanityStrObj(varName, 'variable name')
 
-        op = 'store'
-
         desc = _PythonFP
         if isinstance(val, str):
             desc |= _DirectString
@@ -1037,15 +1177,14 @@ class NetWorkSpace(object):
             except Exception, e:
                 raise NwsPicklingException(e, self.curWs, varName)
 
-        self.send('0005%020d%s%020d%s%020d%s%020d%s%020d' % \
-                (len(op), op, len(self.curWs), self.curWs,
-                 len(varName), varName, len(descTxt), descTxt, len(pVal)))
+        self.server.sendOpStreaming('store', self.curWs, varName, descTxt, **metadata)
+        self.send('%020d' % len(pVal))
         self.send(pVal)
-
+        self.receiveMetadata()
         status = int(self.recv(4))
         if status: raise NwsOperationException('store failed')
 
-    def fetchFile(self, varName, fobj):
+    def fetchFile(self, varName, fobj, metadata={}):
         """Return and remove a value of a variable from a workspace.
 
         ws.fetchFile(varName, fobj) -> number of bytes written to file
@@ -1057,9 +1196,9 @@ class NetWorkSpace(object):
         fobj -- File to write data to.
 
         """
-        return self.__retrieve(varName, 'fetch', fobj=fobj)
+        return self.__retrieve(varName, 'fetch', fobj=fobj, metadata=metadata)
 
-    def fetchTryFile(self, varName, fobj):
+    def fetchTryFile(self, varName, fobj, metadata={}):
         """Try to return and remove a value of a variable from a workspace.
 
         ws.fetchTryFile(varName, fobj) -> number of bytes written to file
@@ -1072,13 +1211,13 @@ class NetWorkSpace(object):
 
         """
         try:
-            return self.__retrieve(varName, 'fetchTry', fobj=fobj)
+            return self.__retrieve(varName, 'fetchTry', fobj=fobj, metadata=metadata)
         except NwsUnpicklingException:
             raise
         except NwsOperationException:
             return False
 
-    def findFile(self, varName, fobj):
+    def findFile(self, varName, fobj, metadata={}):
         """Return a value of a variable from a workspace.
 
         ws.findFile(varName, fobj) -> number of bytes written to file
@@ -1090,9 +1229,9 @@ class NetWorkSpace(object):
         fobj -- File to write data to.
 
         """
-        return self.__retrieve(varName, 'find', fobj=fobj)
+        return self.__retrieve(varName, 'find', fobj=fobj, metadata=metadata)
 
-    def findTryFile(self, varName, fobj):
+    def findTryFile(self, varName, fobj, metadata={}):
         """Try to return a value of a variable in the workspace.
 
         ws.findTryFile(varName, fobj) -> number of bytes written to file
@@ -1105,13 +1244,13 @@ class NetWorkSpace(object):
 
         """
         try:
-            return self.__retrieve(varName, 'findTry', fobj=fobj)
+            return self.__retrieve(varName, 'findTry', fobj=fobj, metadata=metadata)
         except NwsUnpicklingException:
             raise
         except NwsOperationException:
             return False
 
-    def storeFile(self, varName, fobj, n=0):
+    def storeFile(self, varName, fobj, n=0, metadata={}):
         """Store a new value into a variable in the workspace from a file.
 
         ws.storeFile(varName, fobj[, n]) -> number of bytes read from file
@@ -1134,8 +1273,6 @@ class NetWorkSpace(object):
         _sanityStrObj(varName, 'variable name')
         _sanityFileObj(fobj, 'r')
 
-        op = 'store'
-
         desc = _PythonFP | _DirectString
         descTxt = '%020u' % desc
 
@@ -1160,7 +1297,8 @@ class NetWorkSpace(object):
         if n <= 0: return 0
         nread = n  # save this for the return value
 
-        self.send('0005%020d%s%020d%s%020d%s%020d%s%020d' % (len(op), op, len(self.curWs), self.curWs, len(varName), varName, len(descTxt), descTxt, n))
+        self.server.sendOpStreaming('store', self.curWs, varName, descTxt, **metadata)
+        self.send('%020d' % n)
 
         # allow file-like objects to tell us how much data to ask
         # for at a time.  however, we also allow them return either
@@ -1198,6 +1336,7 @@ class NetWorkSpace(object):
                     self.send(n * "\0")
                     break
 
+        self.receiveMetadata()
         status = int(self.recv(4))
         if status: raise NwsOperationException('store file failed')
 
@@ -1206,9 +1345,8 @@ class NetWorkSpace(object):
     def _iretrieve(self, varName, op, varId, valIndex, fobj=None):
         _sanityFileObj(fobj, 'w')
 
-        self.send('0005%020d%s%020d%s%020d%s%020d%-20.20s%020d%020d' % \
-                (len(op), op, len(self.curWs), self.curWs,
-                 len(varName), varName, 20, varId, 20, valIndex))
+        self.server.sendOp(op, self.curWs, varName, "%-20.20s" % varId, "%020d" % valIndex)
+        self.receiveMetadata()
         status = int(self.recv(4)) # barely used at the moment.
 
         # even if failure status, read the rest of the bytes.
@@ -1343,6 +1481,14 @@ class NetWorkSpace(object):
         else:
             raise ValueError('illegal mode specified: %r' % mode)
 
+    def view(self, port=8766):
+        if webbrowser:
+            vstring = "http://%s:%d/doit?op=listVars&wsName=%s" % \
+                    (self.server.serverHost, port, quote_plus(self.curWs))
+            webbrowser.open_new(vstring)
+        else:
+            raise NwsUnsupportedMethodException('cannot import webbrowser module')
+
 class NwsValueIterator(object):
 
     """Implements the iterated operations against a workspace variable.
