@@ -57,6 +57,12 @@ except: platform = None
 try: import subprocess
 except: pass
 
+try:
+    import win32api
+    _winuser = win32api.GetUserName()
+except ImportError:
+    _winuser = None
+
 try:
     from itertools import izip, count
 except ImportError:
@@ -72,14 +78,21 @@ except ImportError:
             yield n
             n += 1
 
+try:
+    import webbrowser
+except ImportError:
+    webbrowser = None
+
 # We use this a lot, so provide convenience name.
 from os import environ as _Env
 from types import CodeType, FunctionType, BuiltinFunctionType, MethodType
 from cStringIO import StringIO
 from warnings import warn, filterwarnings
+from urllib import quote_plus
 
 from nws.client import NetWorkSpace, NwsServer, NwsNoWorkSpaceException
 from nws.client import FIFO, DICT, V_VARIABLE, V_VALUES, V_FETCHERS, V_FINDERS, V_MODE
+from nws.client import PROTOCOL_VERSION_OLD, PROTOCOL_VERSION_NEW_0
 from nws.util import which, msc_quote
 from nws import __version__
 
@@ -92,7 +105,8 @@ __all__ = [
     'SleighGatheredException', 'SleighStoppedException',
     'SleighOccupiedException', 'SleighIllegalValueException',
     'SleighScriptException', 'SleighTaskException',
-    'SleighJoinException', 'SleighNotAllowedException'
+    'SleighJoinException', 'SleighNotAllowedException',
+    'SleighUnsupportedMethodException'
 ]
 
 # We use alternating barriers to synchronize eachWorker
@@ -205,6 +219,9 @@ class SleighJoinException(SleighException):
 class SleighNotAllowedException(SleighException):
     """Sleigh construction not allowed."""
 
+class SleighUnsupportedMethodException(SleighException):
+    """Unsupported method."""
+
 # functions needed for pickling and unpickling code objects
 def _code_ctor(*args):
     return new.code(*args)
@@ -289,6 +306,7 @@ def cmdLaunch(verbose=0):
         'PythonSleighNwsPort': ('nwsPort', int),
         'PythonSleighWorkerCount': ('maxWorkerCount', int),
         'PythonSleighName': ('name', str),
+        'PythonSleighRank': ('rank', int),
         'PythonSleighRedirectOutput': ('redirectOutput', xbool),
     }
     kw = {'verbose': verbose}
@@ -301,7 +319,7 @@ def cmdLaunch(verbose=0):
 
 def launch(nwsName, nwsHost='localhost', nwsPort=8765,
         maxWorkerCount=-1, name=socket.getfqdn(), verbose=0,
-        userNwsName='__default', redirectOutput=True):
+        userNwsName='__default', rank=None, redirectOutput=True):
     """Execute sleigh tasks on behalf of the sleigh owner.
 
     launch(nwsName[, nwsHost, nwsPort, maxWorkerCount, name, verbose, userNwsName, redirectOutput])
@@ -330,6 +348,8 @@ def launch(nwsName, nwsHost='localhost', nwsPort=8765,
             value is the fully qualified domain name of the local
             machine.
 
+    rank -- worker rank, if we're using the new protocol
+
     verbose -- Boolean flag used for displaying debug messages.  The
             default value is 0.
 
@@ -343,17 +363,43 @@ def launch(nwsName, nwsHost='localhost', nwsPort=8765,
     nws = NetWorkSpace(nwsName, nwsHost, nwsPort, useUse=True, create=False)
     userNws = nws.server.useWs(userNwsName, create=False)
 
-    rank = nws.fetch('rankCount')
-    if rank < 0:
-        nws.store('rankCount', rank)
-        raise SleighJoinException('too late to join worker group')
-    elif maxWorkerCount >= 0 and rank + 1 >= maxWorkerCount:
-        # I think this handles a maxWorkerCount of zero, but it
-        # shouldn't happen unless the user calls launch manually.
-        nws.store('workerCount', maxWorkerCount)
-        nws.store('rankCount', -1)
+    if platform:
+        worker_info = {
+                'host': platform.node(),
+                'os': platform.system(),
+                'pid': os.getpid(),
+                'python': platform.python_version(),
+                'nws': __version__,
+                'rank': rank,
+                'logfile': _Env.get('PythonSleighLogFile', '[unknown]')}
     else:
-        nws.store('rankCount', rank + 1)
+        worker_info = {
+                'host': socket.getfqdn(),
+                'os': sys.platform,
+                'pid': os.getpid(),
+                'python': sys.version,
+                'nws': __version__,
+                'rank': rank,
+                'logfile': _Env.get('PythonSleighLogFile', '[unknown]')}
+
+    if rank is not None:
+        new_protocol = True
+        nws.store('worker_ids', str(rank), metadata=worker_info)
+    else:
+        new_protocol = False
+        rank = nws.fetch('rankCount')
+        worker_info['rank'] = rank
+        nws.store('worker_info', worker_info)
+        if rank < 0:
+            nws.store('rankCount', rank)
+            raise SleighJoinException('too late to join worker group')
+        elif maxWorkerCount >= 0 and rank + 1 >= maxWorkerCount:
+            # I think this handles a maxWorkerCount of zero, but it
+            # shouldn't happen unless the user calls launch manually.
+            nws.store('workerCount', maxWorkerCount)
+            nws.store('rankCount', -1)
+        else:
+            nws.store('rankCount', rank + 1)
 
     # initialize for monitoring
     displayName = '%s@%d' % (name, rank)
@@ -366,31 +412,21 @@ def launch(nwsName, nwsHost='localhost', nwsPort=8765,
         nodeList += ' ' + displayName
     nws.store('nodeList', nodeList)
 
-    # post some info about this worker.
-    if platform:
-        nws.store('worker info', {
-                'host': platform.node(),
-                'os': platform.system(),
-                'pid': os.getpid(),
-                'python': platform.python_version(),
-                'nws': __version__,
-                'rank': rank,
-                'logfile': _Env.get('PythonSleighLogFile', '[unknown]')})
-    else:
-        nws.store('worker info', {
-                'host': socket.getfqdn(),
-                'os': sys.platform,
-                'pid': os.getpid(),
-                'python': sys.version,
-                'nws': __version__,
-                'rank': rank,
-                'logfile': _Env.get('PythonSleighLogFile', '[unknown]')})
+    nws.store('workerDebug', displayName + ': Updated node list.')
 
-    # wait for all workers to join
+    # wait for all workers to join (plugin does not wait)
+    # XXX: Should I still do this with the plugin?  I will answer your question
+    #      with another question.  Do you feel like making fglob fetch
+    #      workerCount lazily?
     workerCount = nws.find('workerCount')
 
+    nws.store('workerDebug', displayName + ': Getting ready to prepare to begin calling the worker loop.')
+
     # enter the main worker loop
-    _workerLoop(nws, displayName, rank, workerCount, verbose, userNws, redirectOutput)
+    try:
+        _workerLoop(nws, displayName, rank, workerCount, verbose, userNws, redirectOutput)
+    except:
+        nws.store('workerDebug', displayName + ': Something is rotten, my friends: ' + str(sys.exc_info()))
 
 class Logger(object):
     def __init__(self, ws, name, role):
@@ -445,6 +481,7 @@ def _workerLoop(nws, displayName, rank, workerCount, verbose, userNws, redirectO
             should be redirected to a workspace variable.
 
     """
+    nws.store('workerDebug', displayName + ': Entering worker loop.')
     bx = 0
 
     # monitoring stuff here
@@ -461,6 +498,7 @@ def _workerLoop(nws, displayName, rank, workerCount, verbose, userNws, redirectO
         'SleighWorkerCount': workerCount,
     }
 
+    nws.store('workerDebug', displayName + ': Creating worker logger.')
     logger = Logger(nws, displayName, 'worker')
 
     if redirectOutput:
@@ -473,6 +511,7 @@ def _workerLoop(nws, displayName, rank, workerCount, verbose, userNws, redirectO
     if not _Env.has_key('PythonSleighAllowed'):
         _Env['PythonSleighAllowed'] = 'FALSE'
 
+    nws.store('workerDebug', displayName + ': Beginning task loop.')
     try:
         while 1:
             # update the number of tasks executed
@@ -700,6 +739,26 @@ def sshforwardcmd(machine, **kw):
     else:
         return [ssh, '-n', '-x', '-R', r, machine]
 
+def rwincmd(machine, **kw):
+    socket.gethostbyname(machine)
+    if not machine:
+        raise SleighIllegalValueException('empty string specified in nodeList')
+    scriptDir = getparam('scriptDir', kw, machine, coerce=None)
+    if scriptDir:
+        vbscript = os.path.join(scriptDir, 'rwin.vbs')
+    else:
+        # scriptDir isn't set, so we'll try to guess where rwin.vbs is
+        vbscript = _findScript('rwin.vbs')
+    user = getparam('user', kw, machine, coerce=None)
+    if not user:
+        user = _Env.get('USER') or _Env.get('USERNAME') or _winuser or 'nwsuser'
+    passwd = getparam('passwd', kw, machine, coerce=None)
+    if not passwd:
+        return ['cscript', '//nologo', vbscript, machine, '--']
+    else:
+        return ['cscript', '//nologo', vbscript, machine, '-l', user,
+                '-p', passwd, '--']
+
 def _pathSplit(p):
     s = []
     while True:
@@ -952,18 +1011,48 @@ defaultSleighOptions = {
     'nodeList':         ['localhost', 'localhost', 'localhost'],
     'scriptExec':       (os.name == 'posix') and envcmd or scriptcmd,
     'scriptDir':        None,
-    'scriptName':       (os.name == 'posix') and \
-            'PythonNWSSleighWorker' or 'PythonNWSSleighWorker.py',
+    'scriptName':       'PythonNWSSleighWorker.py',
     'modulePath':       None,
     'workingDir':       os.getcwd(),
     'logDir':           None,
     'user':             None,
+    'passwd':           None,
     'wsNameTemplate':   'sleigh_ride_%04d',
     'userWsNameTemplate': 'sleigh_user_%04d',
     'verbose':          False,
     'redirectOutput':   True,
 }
 
+class Barrier(object):
+    def __init__(self, ws, name):
+        self.ws   = ws
+        self.name = name
+
+    def leave(self, block=False):
+        if block:
+            self.ws.fetch(self.name)
+        else:
+            self.ws.fetchTry(self.name)
+
+    def enter(self):
+        self.ws.store(self.name, '1')
+
+    def wait(self):
+        self.ws.find(self.name)
+
+class NullBarrier(object):
+    def __init__(self):
+        pass
+
+    def leave(self, block=False):
+        pass
+
+    def enter(self):
+        pass
+
+    def wait(self):
+        pass
+
 class Accumulator(object):
     def __init__(self):
         self._val = {}
@@ -995,7 +1084,7 @@ class SleighPending(object):
     the results of the operation to be returned.
     """
 
-    def __init__(self, nws, numTasks, barrierName, sleighState, accumulator):
+    def __init__(self, nws, id, numTasks, barrier, my_sleigh, accumulator):
         """Create an object that represents the pending sleigh operation.
 
         This constructor is intended for internal use only.
@@ -1015,9 +1104,10 @@ class SleighPending(object):
 
         """
         self.nws = nws
+        self.id = id
         self.numTasks = numTasks
-        self.barrierName = barrierName
-        self.sleighState = sleighState
+        self.barrier = barrier
+        self.my_sleigh = my_sleigh
         self.accumulator = accumulator
         self.done = False
 
@@ -1053,16 +1143,11 @@ class SleighPending(object):
         if self.done:
             raise SleighGatheredException('results already gathered')
 
-        for i in xrange(self.numTasks):
-            r = self.nws.fetch('result')
-            if r.error:
-                _accumerror(self.accumulator, [(r.tag, r.value)])
-            else:
-                _accumresult(self.accumulator, [(r.tag, r.value)])
+        self.sleigh.__retrieveResults(self.id, self.numTasks, self.accumulator)
 
-        if self.barrierName: self.nws.store(self.barrierName, 1)
+        if self.barrier is not None: self.barrier.enter()
 
-        self.sleighState.occupied = False
+        self.sleigh.sleighState.occupied = False
         self.done = True
 
         return _accumvalue(self.accumulator)
@@ -1256,7 +1341,7 @@ class Sleigh(object):
                     % (opts['nwsHost'], e.args[1]))
 
         # set up the sleigh's netWorkSpace.
-        self.nwss = NwsServer(host=opts['nwsHost'], port=opts['nwsPort'])
+        self.nwss = NwsServer(host=opts['nwsHost'], port=opts['nwsPort'], connopts={'MetadataToServer': '1', 'MetadataFromServer': '1'})
 
         # last minute sanity checking of nodeList when using service launch
         if opts['launch'] == 'service':
@@ -1282,7 +1367,7 @@ class Sleigh(object):
                     raise SleighIllegalValueException('bad node %s in nodeList: %s'
                             % (node, 'no such worker registered'))
 
-        self.nwsName = self.nwss.mktempWs(opts['wsNameTemplate'])
+        self.nwsName = self.nwss.mktempWs(opts['wsNameTemplate'], metadata={'wstype': 'sleigh'})
         if not self.nwsName:
             self.nwss.close()
             raise SleighIllegalValueException('bad wsNameTemplate value "%s": %s'
@@ -1296,16 +1381,16 @@ class Sleigh(object):
                     % (opts['nwsHost'], e.args[1]))
         self.userNws = self.nwss.openWs(self.userNwsName)
 
-        # initialize for monitoring
-        self.nws.declare('nodeList', 'single')
-        self.nws.store('nodeList', '')
-        self.nws.declare('totalTasks', 'single')
-        self.nws.store('totalTasks', '0')
+        self.__initialize(self.nwss.get_protocol_version())
+        self.__launch_workers(opts)
 
-        self.nws.declare('rankCount', 'single')
-        self.nws.store('rankCount', 0)
-        self.nws.declare('workerCount', 'single')
+        self.state =  _SleighState()
+        self.state.bx = 0
+        self.state.occupied = False
+        self.state.stopped = False
+        self.state.totalTasks = 0
 
+    def __launch_workers(self, opts):
         if isinstance(opts['launch'], FunctionType):
             # remote launch
             self.nodeList = opts['nodeList']
@@ -1377,10 +1462,11 @@ class Sleigh(object):
 
             self.nws.deleteVar('runMe')
             self.workerCount = self.nws.fetch('rankCount')
-            self.nws.store('workerCount', self.workerCount)
-            self.nws.store('rankCount', -1)
-            self.rankCount = -1
-            self.nodeList = self.nws.fetchTry('nodeList', '')
+            if not self.new_protocol:
+                self.nws.store('workerCount', self.workerCount)
+                self.nws.store('rankCount', -1)
+                self.rankCount = -1
+                self.nodeList = self.nws.fetchTry('nodeList', '')
         else:
             # the previous error checking should prevent this from
             # ever happening
@@ -1389,11 +1475,80 @@ class Sleigh(object):
             self.rankCount = -1
             self.nodeList = ''
 
-        self.state =  _SleighState()
-        self.state.bx = 0
-        self.state.occupied = False
-        self.state.stopped = False
-        self.state.totalTasks = 0
+    def __initialize(self, proto_version):
+        v = self.nws.findTry('version')
+        if v is None:
+            proto_version = PROTOCOL_VERSION_OLD
+
+        if proto_version == PROTOCOL_VERSION_OLD:
+            self.__initialize_old()
+        elif proto_version == PROTOCOL_VERSION_NEW_0:
+            print 'New protocol'
+            self.__initialize_new_0()
+        else:
+            raise Exception('Unknown protocol version')
+
+    def __initialize_old(self):
+        self.new_protocol = False
+
+        # initialize for monitoring
+        self.nws.declare('nodeList', 'single')
+        self.nws.store('nodeList', '')
+        self.nws.declare('totalTasks', 'single')
+        self.nws.store('totalTasks', '0')
+
+        self.nws.declare('rankCount', 'single')
+        self.nws.store('rankCount', 0)
+        self.nws.declare('workerCount', 'single')
+
+        self.__statusImpl = self.__status_old
+        self._syncWorkerCount = self.__syncWorkerCount_old
+        self.__storeBroadcastTask = self.__storeBroadcastTask_old
+        self.__newBatch = lambda: ''
+        self.__barriers = [Barrier(self.nws, _barrierNames[i]) for i in range(2)]
+
+    def __initialize_new_0(self):
+        self.nws.declare('nodeList', 'single')
+        self.nws.store('nodeList', '')
+        self.__nextBatch = 0
+        self.new_protocol = True
+        if self.options['launch'] != 'web':
+            self.nws.store('workerCount', '%d' % len(self.options['nodeList']))
+        self.__statusImpl = self.__status_new_0
+        self._syncWorkerCount = self.__syncWorkerCount_new_0
+        self.__storeBroadcastTask = self.__storeBroadcastTask_new_0
+        self.__newBatch = self.__newBatch_new
+        self.__barriers = [NullBarrier() for i in range(2)]
+
+    def __newBatch_new(self):
+        id = '%d' % self.__nextBatch
+        self.__nextBatch += 1
+        return id
+
+    def __storeBroadcastTask_old(self, wc, task):
+        # update the total number of submitted tasks
+        self.state.totalTasks += wc
+        self.nws.store('totalTasks', str(self.state.totalTasks))
+
+        for i in xrange(wc):
+            task.tag = i
+            self.nws.store('task', task)
+
+        # no batch ids in old versions
+        return ''
+
+    def __storeBroadcastTask_new_0(self, wc, task):
+        id = self.__newBatch()
+        self.nws.store('broadcast', task, metadata={'batchId': id})
+        return id
+
+    def __retrieveResults(self, id, ntasks, accum):
+        for i in xrange(ntasks):
+            r = self.nws.fetch('result', metadata={'batchId': id})
+            if r.error:
+                _accumerror(accum, [(r.tag, r.value)])
+            else:
+                _accumresult(accum, [(r.tag, r.value)])
 
     def __str__(self):
         return "Sleigh [%s] on nodes: %s" % \
@@ -1507,51 +1662,68 @@ class Sleigh(object):
             # join phase completed before
             return self.workerCount, 1
         else:
-            sleepTime = initialSleep = min(2.0, timeout)
-            repeatedSleep = 1.0
-            startTime = time.time()
+            return self.__statusImpl(closeGroup, timeout)
 
-            while 1:
-                n = self.nws.fetch('rankCount')
-                if n < 0:
-                    # all workers joined
-                    if self.options['verbose']:
-                        print 'all %d workers have started' % self.workerCount
+    def __status_new_0(self, closeGroup, timeout):
+        if closeGroup:
+            self.workerCount = self.nws.fetch('status', metadata={'delay': str(timeout)})
+            self.rankCount = -1
+            return self.workerCount, 1
+        else:
+            status = self.nws.fetch('join_status')
+            if status == 'closed':
+                self.workerCount = self.nws.fetch('status', metadata={'delay': str(timeout)})
+                self.rankCount = -1
+                return self.workerCount, 1
+            else:
+                return self.nws.fetch('waitForWorkers', metadata={'delay': str(timeout)}), 0
 
-                    assert self.workerCount == self.nws.find('workerCount')
-                    self.rankCount = -1
-                    self.nws.store('rankCount', self.rankCount)
-                    return self.workerCount, 1
-                else:
-                    # three choices: close now, return not closed, or
-                    # sleep and try again
-                    if time.time() - startTime >= timeout:
-                        # time is up, so either close the join, or
-                        # return the current status
-                        if closeGroup:
-                            if self.options['verbose']:
-                                print 'closing group: %d workers' % n
+    def __status_old(self, closeGroup, timeout):
+        sleepTime = initialSleep = min(2.0, timeout)
+        repeatedSleep = 1.0
+        startTime = time.time()
 
-                            self.workerCount = n
-                            self.nws.store('workerCount', self.workerCount)
-                            self.rankCount = -1
-                            self.nws.store('rankCount', self.rankCount)
-                            return self.workerCount, 1
-                        else:
-                            if self.options['verbose']:
-                                print 'group not formed: %d workers' %n
+        while 1:
+            n = self.nws.fetch('rankCount')
+            if n < 0:
+                # all workers joined
+                if self.options['verbose']:
+                    print 'all %d workers have started' % self.workerCount
 
-                            self.rankCount = n
-                            self.nws.store('rankCount', self.rankCount)
-                            return self.rankCount, 0
+                assert self.workerCount == self.nws.find('workerCount')
+                self.rankCount = -1
+                self.nws.store('rankCount', self.rankCount)
+                return self.workerCount, 1
+            else:
+                # three choices: close now, return not closed, or
+                # sleep and try again
+                if time.time() - startTime >= timeout:
+                    # time is up, so either close the join, or
+                    # return the current status
+                    if closeGroup:
+                        if self.options['verbose']:
+                            print 'closing group: %d workers' % n
+
+                        self.workerCount = n
+                        self.nws.store('workerCount', self.workerCount)
+                        self.rankCount = -1
+                        self.nws.store('rankCount', self.rankCount)
+                        return self.workerCount, 1
                     else:
                         if self.options['verbose']:
-                            print 'only %d workers: sleeping...' % n
+                            print 'group not formed: %d workers' %n
 
                         self.rankCount = n
-                        self.nws.store('rankCount', n)
-                        time.sleep(sleepTime)
-                        sleepTime = repeatedSleep
+                        self.nws.store('rankCount', self.rankCount)
+                        return self.rankCount, 0
+                else:
+                    if self.options['verbose']:
+                        print 'only %d workers: sleeping...' % n
+
+                    self.rankCount = n
+                    self.nws.store('rankCount', n)
+                    time.sleep(sleepTime)
+                    sleepTime = repeatedSleep
 
     def stop(self):
         """Stop the remote processes and delete the sleigh workspace.
@@ -1576,10 +1748,17 @@ class Sleigh(object):
         self.nwss.close()
         self.state.stopped = True
 
-    def _syncWorkerCount(self):
-        self.rankCount = self.nws.find('rankCount')
-        if self.rankCount == -1:
-            self.workerCount = self.nws.find('workerCount')
+    def __syncWorkerCount_old(self):
+        if self.rankCount != -1:
+            self.rankCount = self.nws.find('rankCount')
+            if self.rankCount == -1:
+                self.workerCount = int(self.nws.find('workerCount'))
+        return self.workerCount
+
+    def __syncWorkerCount_new_0(self):
+        if self.rankCount != -1:
+            self.workerCount = int(self.nws.find('workerCount'))
+        return self.workerCount
 
     def eachWorker(self, fun, *workerArgs, **kw):
         """Execute a function, method, or expression on each worker of sleigh.
@@ -1671,40 +1850,25 @@ class Sleigh(object):
 
         t = _newtask(fun, taskType, True)
 
-        if self.rankCount != -1:
-            self._syncWorkerCount()
+        self._syncWorkerCount()
 
         nws = self.nws
         wc = self.workerCount
 
         # use alternating barrier to sync eachWorker invocations with the workers.
-        bn = _barrierNames[self.state.bx]
+        cur_barrier = self.__barriers[self.state.bx]
         self.state.bx ^= 1
+        cur_barrier.leave()
 
-        nws.fetchTry(bn)
-
-        # update the total number of submitted tasks
-        self.state.totalTasks += wc
-        nws.store('totalTasks', str(self.state.totalTasks))
-
-        for i in xrange(wc):
-            t.tag = i
-            t.args = list(workerArgs)
-            nws.store('task', t)
+        t.args = list(workerArgs)
+        id = self.__storeBroadcastTask(wc, t)
 
         if not blocking:
-            self.state.occupied = True
-            return SleighPending(nws, wc, bn, self.state, accumulator)
-
-        for i in xrange(wc):
-            r = nws.fetch('result')
-            if r.error:
-                _accumerror(accumulator, [(r.tag, r.value)])
-            else:
-                _accumresult(accumulator, [(r.tag, r.value)])
-
-        nws.store(bn, 1)
+            self.state.occupied = not self.new_protocol
+            return SleighPending(nws, id, wc, cur_barrier, self, accumulator)
 
+        self.__retrieveResults(id, wc, accumulator)
+        cur_barrier.enter()
         return _accumvalue(accumulator)
 
     # run fun once for each element of a vector.
@@ -1888,45 +2052,51 @@ class Sleigh(object):
 
         t = _newtask(fun, taskType, False)
 
-        if self.rankCount != -1:
-            self._syncWorkerCount()
-
         nws = self.nws
-        wc = self.workerCount
 
         # create an "enumerator" for the task data
         taskIter = izip(count(), *elementArgs)
 
+        # allocate a batch id for the tasks
+        id = self.__newBatch()
+        md = {'batchId': id}
+
         # loadFactor is ignored if non-blocking
         if blocking and loadFactor > 0:
+            wc = self._syncWorkerCount()
             taskLimit = xrange(max(loadFactor * wc, wc))
         else:
             taskLimit = count()  # no limit
 
+        # task args generator
+        argsgen = ((taskData[0], (list(taskData[1:])+fixedArgs)) for taskData in taskIter)
+        if argPermute is not None:
+            argsgen = ((td[0], [td[1][j] for j in argPermute]) for td in argsgen)
+
         # submit tasks, possibly limited by loadFactor
         initialSubmission = 0
-        for ignore, taskData in izip(taskLimit, taskIter):
+        for ignore, taskData in izip(taskLimit, argsgen):
             t.tag = taskData[0]
-            t.args = list(taskData[1:]) + fixedArgs
-            if argPermute is not None: t.args = [t.args[j] for j in argPermute]
-            nws.store('task', t)
+            t.args = taskData[1]
+            nws.store('task', t, md)
             initialSubmission += 1
 
         # update the total number of submitted tasks
-        self.state.totalTasks += initialSubmission
-        nws.store('totalTasks', str(self.state.totalTasks))
+        if not self.new_protocol:
+            self.state.totalTasks += initialSubmission
+            nws.store('totalTasks', str(self.state.totalTasks))
 
         # return a SleighPending object if we're non-blocking
         if not blocking:
-            self.state.occupied = True
-            return SleighPending(nws, initialSubmission, '', self.state, accumulator)
+            self.state.occupied = not self.new_protocol
+            return SleighPending(nws, id, initialSubmission, None, self, accumulator)
 
         # start retrieving results and finish submitting tasks
         # which is needed when loadFactor > 0
         numTasks = initialSubmission
-        for taskData in taskIter:
+        for taskData in argsgen:
             # fetch a result
-            r = nws.fetch('result')
+            r = nws.fetch('result', md)
             if r.error:
                 _accumerror(accumulator, [(r.tag, r.value)])
             else:
@@ -1934,19 +2104,19 @@ class Sleigh(object):
 
             # submit next task
             t.tag = taskData[0]
-            t.args = list(taskData[1:]) + fixedArgs
-            if argPermute is not None: t.args = [t.args[j] for j in argPermute]
-            nws.store('task', t)
+            t.args = taskData[1]
+            nws.store('task', t, md)
             numTasks += 1
 
             # update the total number of submitted tasks
-            self.state.totalTasks += 1
-            nws.store('totalTasks', str(self.state.totalTasks))
+            if not self.new_protocol:
+                self.state.totalTasks += 1
+                nws.store('totalTasks', str(self.state.totalTasks))
 
         # finish retrieving results
         for i in xrange(initialSubmission):
             # fetch a result
-            r = nws.fetch('result')
+            r = nws.fetch('result', md)
             if r.error:
                 _accumerror(accumulator, [(r.tag, r.value)])
             else:
@@ -2023,17 +2193,14 @@ class Sleigh(object):
 
         t = _newtask(fun, taskType, False)
 
-        if self.rankCount != -1:
-            self._syncWorkerCount()
-
         nws = self.nws
-        wc = self.workerCount
 
         # create an "enumerator" for the task data
         taskIter = izip(count(), *iterables)
 
         # loadFactor is ignored if non-blocking
         if loadFactor > 0:
+            wc = self._syncWorkerCount()
             taskLimit = xrange(max(loadFactor * wc, wc))
         else:
             taskLimit = count()  # no limit
@@ -2120,17 +2287,14 @@ class Sleigh(object):
 
         t = _newtask(fun, taskType, False)
 
-        if self.rankCount != -1:
-            self._syncWorkerCount()
-
         nws = self.nws
-        wc = self.workerCount
 
         # create an "enumerator" for the task data
         taskIter = izip(count(), iterable)
 
         # loadFactor is ignored if non-blocking
         if loadFactor > 0:
+            wc = self._syncWorkerCount()
             taskLimit = xrange(max(loadFactor * wc, wc))
         else:
             taskLimit = count()  # no limit
@@ -2177,6 +2341,22 @@ class Sleigh(object):
         aux.sort()
         return [info[i] for rank, i in aux]
 
+    def view(self, port=8766, ws='system'):
+        if webbrowser:
+            if ws == 'system':
+                nws = self.nws
+            elif ws == 'user':
+                nws = self.userNws
+            else:
+                raise SleighIllegalValueException(
+                        'illegal value specified for "ws" argument: %s' % ws)
+
+            vstring = "http://%s:%d/doit?op=listVars&wsName=%s" % \
+                    (nws.server.serverHost, port, quote_plus(nws.curWs))
+            webbrowser.open_new(vstring)
+        else:
+            raise SleighUnsupportedMethodException('cannot import webbrowser module')
+
 class SleighResultIterator(object):
 
     """Returns results from tasks submitted to the sleigh.
