@@ -20,6 +20,7 @@
 
 import sys, os, tempfile, traceback, socket, time, subprocess
 from os import environ as Env
+from nws.client import NetWorkSpace
 
 try:
     import win32api
@@ -84,6 +85,12 @@ def main():
         logger.setName(name)
     logger.logDebug(nwsName, nwsHost, nwsPort)
 
+    nws = NetWorkSpace(nwsName, nwsHost, nwsPort, useUse=True, create=False)
+    newProtocol = nws.findTry('version') is not None
+    if newProtocol:
+        worker_id = nws.fetch('worker_ids')
+        Env['PythonSleighRank'] = worker_id
+
     # create the script file for the worker to execute
     script = '''\
 import sys, os
@@ -121,8 +128,6 @@ cmdLaunch(%d)
 
     logger.logDebug("waiting for shutdown")
     sys.path[1:1] = modulePath.split(os.pathsep)
-    from nws.client import NetWorkSpace
-    nws = NetWorkSpace(nwsName, nwsHost, nwsPort, useUse=True, create=False)
     logger.setWs(nws)
     sentinel(nws)
 
