@@ -7,7 +7,7 @@ from math import sqrt
 
 # start the local nws server
 srv = NwsLocalServer(daemon=False)
-srv.waitUntilStarted()
+srv.wait_until_started()
 
 # try an nws example
 ws = NetWorkSpace('foo', serverPort=srv.port)
